"""Various functions for edge profiles."""

__all__ = [
    "preprocess",
    "fill_after",
]


import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks
from scipy.stats import linregress


def preprocess(Ys, sigma, std_thres):
    """Preprocess raw profiles.

    Parameters
    ----------
    Ys : (N, M) array
        Array of N profiles.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.
    std_thres : scalar
        Standard deviation threshold to detect contact point.

    Returns
    -------
    Ys : (N, M) array
        Preprocessed profile data.
    Ls : (N,) array
        Length of *Y* until the contact point.

    Notes
    -----
    Profiles undergo the following steps:

    1. Profile direction is set so that the contact point is on the right hand side.
    2. Contact point is detected, and set to have zero height.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, RawProfileCsvs
    >>> from heavyedge.profile import preprocess
    >>> raw = RawProfileCsvs(get_sample_path("Type3"))
    >>> Ys = np.array([raw[i][0] for i in range(len(raw))])
    >>> Ys_processed, Ls = preprocess(Ys, 32, 0.01)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for Y, L in zip(Ys_processed, Ls):
    ...     plt.plot(Y[:L])
    """
    ret_Ys, ret_Ls = [], []
    for Y in Ys:
        Y_proc, L = _preprocess(Y, sigma, std_thres)
        ret_Ys.append(Y_proc)
        ret_Ls.append(L)
    return np.array(ret_Ys), np.array(ret_Ls)


def _preprocess(Y, sigma, std_thres):
    if Y[0] < Y[-1]:
        # Make plateau is on the left and cp is on the right
        Y = np.flip(Y)

    X = np.arange(len(Y))
    h_xx = gaussian_filter1d(Y, sigma, order=2, mode="nearest")
    if len(h_xx) > 0:
        peaks, _ = find_peaks(h_xx)
    else:
        peaks = np.empty(0, dtype=int)

    candidates = []
    for i, peak_idx in enumerate(peaks):
        x = X[peak_idx:]
        if not len(x) > 2:
            continue
        y = Y[x]
        reg = linregress(x, y)
        residuals = y - (reg.intercept + reg.slope * x)
        std = np.sqrt(np.sum(residuals**2) / (len(x) - 2))
        if std < std_thres:
            candidates.append(i)

    if candidates:
        cp = peaks[candidates[np.argmax(h_xx[peaks[candidates]])]]
    else:
        cp = len(Y) - 1

    # If any point before cp is lower than the detected contact point,
    # set that as contact point instead.
    cp = np.argmin(Y[: cp + 1])
    Y = Y - Y[cp]
    return Y, cp + 1


def fill_after(Ys, Ls, fill_value):
    """Fill arrays with a constant value after specified lengths.

    The input array *Ys* is modified.

    Parameters
    ----------
    Ys : (N, M) array
        Array of N profiles.
    Ls : (N,) array
        Length of each profile.
    fill_value : scalar
        Value to fill *Ys*.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.profile import fill_after
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> fill_after(Ys, Ls, float("nan"))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T)
    """
    _, M = Ys.shape
    Ys[np.arange(M)[None, :] >= Ls[:, None]] = fill_value



"""Commands to process profiles."""

import pathlib
from importlib.metadata import entry_points

import numpy as np

from heavyedge.cli.command import Command, deprecate_command, register_command

PLUGIN_ORDER = 0.0


@register_command("prep", "Preprocess raw profiles")
class PrepCommand(Command):
    def add_parser(self, main_parser):
        prep = main_parser.add_parser(
            self.name,
            description="Preprocess raw profiles and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        prep.add_argument(
            "--type",
            choices=[ep.name for ep in entry_points(group="heavyedge.rawdata")],
            required=True,
            help="Type of the raw profile data.",
        )
        prep.add_argument("--name", help="Name to label output dataset.")
        prep.add_argument(
            "raw",
            type=pathlib.Path,
            help="Path to raw profile data.",
        )
        prep.add_config_argument(
            "--res",
            type=float,
            help=(
                "Spatial resolution of profile data, i.e., points per unit length. "
                "Length unit must match the profile height unit."
            ),
        )
        prep.add_config_argument(
            "--sigma",
            type=float,
            help="Standard deviation of Gaussian kernel.",
        )
        prep.add_config_argument(
            "--std-thres",
            type=float,
            help="Standard deviation threshold for contact point detection.",
        )
        prep.add_config_argument(
            "--fill-value",
            choices=["0", "nan"],
            help="Value to fill profile after the contact point.",
        )
        prep.add_config_argument(
            "--z-thres",
            type=float,
            help=(
                "Modified Z-score threshold for outlier detection. "
                "If not passed, outliers are not detected."
            ),
        )
        prep.add_config_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, load entire profiles.",
        )
        prep.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.api import prep
        from heavyedge.io import ProfileData

        self.logger.info(f"Writing {args.output}")

        raw_type = entry_points(group="heavyedge.rawdata")[args.type].load()
        raw = raw_type(args.raw)

        if args.fill_value == "0":
            args.fill_value = 0

        gen = prep(
            raw,
            args.sigma,
            args.std_thres,
            args.fill_value,
            args.z_thres,
            args.batch_size,
            lambda msg: self.logger.info(f"{args.output} : {msg}"),
        )

        # Get first result to determine M
        Ys, Ls, names = next(gen)
        M = len(Ys[0])
        with ProfileData(args.output, "w").create(M, args.res, args.name) as out:
            out.write_profiles(Ys, Ls, names)
            for Ys, Ls, names in gen:
                out.write_profiles(Ys, Ls, names)

        self.logger.info(f"Saved {out.path}")

    @staticmethod
    def is_invalid(profile):
        return (len(profile) == 0) or np.any(np.isnan(profile))


@deprecate_command("1.6", "filtering in preparation step (heavyedge prep command)")
@register_command("outlier", "Filter outlier profiles")
class OutlierCommand(Command):
    def add_parser(self, main_parser):
        outlier = main_parser.add_parser(
            self.name,
            description="Remove outlier profiles and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        outlier.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        outlier.add_config_argument(
            "--z",
            type=float,
            help="Modified Z-score threshold for outlier detection.",
        )
        outlier.add_argument(
            "-o", "--output", type=pathlib.Path, help="Output file path"
        )

    def run(self, args):
        from heavyedge import ProfileData
        from heavyedge.api import outlier

        self.logger.info(f"Removing outliers: {args.profiles}")

        with ProfileData(args.profiles) as data:
            _, M = data.shape()
            res = data.resolution()
            name = data.name()

            x = data.x()
            areas = []
            for i in range(len(data)):
                Y, L, _ = data[i]
                areas.append(np.trapezoid(Y[:L], x[:L]))
            is_outlier = outlier(np.array(areas), args.z)

            with ProfileData(args.output, "w").create(M, res, name) as out:
                for skip, Y, L, name in zip(
                    is_outlier,
                    data._file["profiles"],
                    data._file["len"],
                    data.profile_names(),
                ):
                    if not skip:
                        out.write_profiles(Y.reshape(1, -1), [L], [name])

        self.logger.info(f"Saved {out.path}")


@register_command("merge", "Merge profile data")
class MergeCommand(Command):
    def add_parser(self, main_parser):
        merge = main_parser.add_parser(
            self.name,
            description="Merge profile data and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        merge.add_argument(
            "profiles",
            nargs="+",
            type=pathlib.Path,
            help="Paths to preprocessed profile data in 'ProfileData' structure.",
        )
        merge.add_argument("--name", help="Name to label output dataset.")
        merge.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, load entire profiles.",
        )
        merge.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.io import ProfileData

        self.logger.info(f"Writing {args.output}")

        with ProfileData(args.profiles[0]) as data:
            _, M = data.shape()
            res = data.resolution()

        with ProfileData(args.output, "w").create(M, res, args.name) as out:
            for p in args.profiles:
                with ProfileData(p) as data:
                    if args.batch_size is not None:
                        for i in range(0, data.shape[0], args.batch_size):
                            out.write_profiles(*data[i : i + args.batch_size])
                    else:
                        out.write_profiles(*data[:])

        self.logger.info(f"Saved {out.path}")


@register_command("filter", "Filter profile data")
class FilterCommand(Command):
    def add_parser(self, main_parser):
        filter_parser = main_parser.add_parser(
            self.name,
            description="Filter profile data and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        filter_parser.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        filter_parser.add_argument(
            "index",
            type=pathlib.Path,
            help="Path to index npy file for filtering.",
        )
        filter_parser.add_argument("--name", help="Name to label output dataset.")
        filter_parser.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, load entire profiles.",
        )
        filter_parser.add_argument(
            "-o", "--output", type=pathlib.Path, help="Output file path"
        )

    def run(self, args):
        from heavyedge.io import ProfileData

        self.logger.info(f"Writing {args.output}")

        index = np.load(args.index)
        N = len(index)

        with ProfileData(args.profiles) as data:
            _, M = data.shape()
            res = data.resolution()

            with ProfileData(args.output, "w").create(M, res, args.name) as out:
                if args.batch_size is not None:
                    for i in range(0, N, args.batch_size):
                        out.write_profiles(*data[index[i : i + args.batch_size]])
                else:
                    out.write_profiles(*data[index])

        self.logger.info(f"Saved {out.path}")



import os
import subprocess

import numpy as np

from heavyedge import ProfileData


def test_process_commands(tmp_rawdata_type2_path, tmp_path):
    processed_path = tmp_path / "ProcessedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "prep",
            "--type",
            "csvs",
            "--res=1",
            "--sigma=1",
            "--std-thres=40",
            "--fill-value=0",
            "--z-thres=3.5",
            tmp_rawdata_type2_path,
            "-o",
            processed_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(processed_path)

    merged_path = tmp_path / "MergedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "merge",
            processed_path,
            processed_path,
            "-o",
            merged_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(merged_path)

    with ProfileData(processed_path) as data:
        N = len(data)
    index_path = tmp_path / "index.npy"
    np.save(index_path, np.arange(N))
    filtered_path = tmp_path / "FilteredProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "filter",
            processed_path,
            index_path,
            "-o",
            filtered_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(filtered_path)


def test_mean_command(tmp_prepdata_type2_path, tmp_path):
    mean_path = tmp_path / "MeanProfile.h5"
    subprocess.run(
        [
            "heavyedge",
            "mean",
            "--wnum",
            "100",
            tmp_prepdata_type2_path,
            "-o",
            mean_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(mean_path)


def test_edge_command(tmp_prepdata_type2_path, tmp_path):
    area_scaled_path = tmp_path / "AreaScaledProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "scale",
            tmp_prepdata_type2_path,
            "--type=area",
            "-o",
            area_scaled_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(area_scaled_path)

    plateau_scaled_path = tmp_path / "PlateauScaledProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "scale",
            tmp_prepdata_type2_path,
            "--type=plateau",
            "-o",
            plateau_scaled_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(plateau_scaled_path)

    trimmed_path = tmp_path / "TrimmedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "trim",
            tmp_prepdata_type2_path,
            "-o",
            trimmed_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(trimmed_path)

    padded_path = tmp_path / "PaddedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "pad",
            tmp_prepdata_type2_path,
            "-o",
            padded_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(padded_path)



"""
Wasserstein distance
--------------------

Wasserstein-related functions.
"""

import warnings

import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import interp1d

from ._wasserstein import _optimize_q, _quantile

__all__ = [
    "quantile",
    "wdist",
    "wmean",
]


def _deprecated(version, replace):
    removed_version = str(int(version.split(".")[0]) + 1) + ".0"

    def decorator(func):
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"{func.__name__}() is deprecated since HeavyEdge {version} "
                f"and will be removed in {removed_version}. "
                f"Use {replace} instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def quantile(x, fs, Ls, t):
    """Convert probability distributions to quantile functions.

    Parameters
    ----------
    x : (M1,) ndarray
        Coordinates of grids over which *fs* are measured.
    fs : (N, M1) ndarray
        Empirical probability density functions.
    Ls : (N,) ndarray
        Length of supports of each *fs*.
    t : (M2,) ndarray
        Points over which the quantile function will be measured.
        Must be strictly increasing from 0 to 1.

    Returns
    -------
    (N, M2) ndarray
        Quantile functions* over *t*.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.wasserstein import quantile
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> fs = Ys / np.trapezoid(Ys, x, axis=-1)[:, np.newaxis]
    >>> t = np.linspace(0, 1, 100)
    >>> Qs = quantile(x, fs, Ls, t)
    """
    Gs = cumulative_trapezoid(fs, x, initial=0, axis=-1)
    return _quantile(x, Gs, Ls.astype(np.int32), t)


def _quantile_old(x, f, t):
    G = cumulative_trapezoid(f, x, initial=0)
    return interp1d(G, x, bounds_error=False, fill_value=(x[0], x[-1]))(t)


@_deprecated("1.6", "HeavyEdge-Distance package")
def wdist(x1, f1, x2, f2, grid_num):
    r"""Wasserstein distance between two 1D probability distributions.

    .. deprecated:: 1.6
        This function will be removed in HeavyEdge 2.0.
        Use HeavyEdge-Distance package instead.

    .. math::

        d_W(f_1, f_2)^2 = \int^1_0 (Q_1(t) - Q_2(t))^2 dt

    where :math:`Q_i` is the quantile function of :math:`f_i`.

    Parameters
    ----------
    x1, f1 : ndarray
        The first empirical probability density function.
    x2, f2 : ndarray
        The second empirical probability density function.
    grid_num : int
        Number of sample points in [0, 1] to approximate the integral.

    Returns
    -------
    scalar
        Wasserstein distance.
    """
    grid = np.linspace(0, 1, grid_num)
    Q1 = _quantile_old(x1, f1, grid)
    Q2 = _quantile_old(x2, f2, grid)
    return np.trapezoid((Q1 - Q2) ** 2, grid) ** 0.5


def wmean(x, fs, Ls, t):
    """Fréchet mean of probability distrubutions using Wasserstein metric.

    Parameters
    ----------
    x : (M1,) ndarray
        Coordinates of grids over which *fs* are measured.
    fs : (N, M1) ndarray
        Empirical probability density functions.
    Ls : (N,) ndarray
        Length of supports of each *fs*.
    t : (M2,) ndarray
        Points over which the quantile function will be measured.
        Must be strictly increasing from 0 to 1.

    Returns
    -------
    f_mean : ndarray
        Fréchet mean of *fs* over *x*.
    L : int
        Length of the support of *f_mean*.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.wasserstein import wmean
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> fs = Ys / np.trapezoid(Ys, x, axis=-1)[:, np.newaxis]
    >>> f_mean, L = wmean(x, fs, Ls, np.linspace(0, 1, 100))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, fs.T, "--", color="gray")
    ... plt.plot(x[:L], f_mean[:L])
    """
    Qs = quantile(x, fs, Ls, t)
    g = np.mean(Qs, axis=0)
    return _wmean(x, t, g)


def _wmean(x, t, g):
    if np.all(np.diff(g) >= 0):
        q = g
    else:
        q = _optimize_q(g)
    pdf = 1 / np.gradient(q, t)
    pdf[-1] = 0
    pdf /= np.trapezoid(pdf, q)

    L = np.searchsorted(x, q[-1]) + 1
    return np.interp(x, q, pdf, left=pdf[0], right=0), L


def _wmean_old(xs, fs, grid_num):
    grid = np.linspace(0, 1, grid_num)
    Q = np.array([_quantile_old(x, f, grid) for x, f in zip(xs, fs)])
    g = np.mean(Q, axis=0)
    if np.all(np.diff(g) >= 0):
        q = g
    else:
        q = _optimize_q(g)
    pdf = 1 / np.gradient(q, grid)
    pdf[-1] = 0
    pdf /= np.trapezoid(pdf, q)
    return q, pdf



"""Estimate average profiles."""

import numpy as np

from heavyedge.wasserstein import _wmean, quantile, wmean

__all__ = [
    "mean_euclidean",
    "mean_wasserstein",
]


def mean_euclidean(f, batch_size=None, logger=lambda x: None):
    """Compute arithmetic mean profile.

    Parameters
    ----------
    f : heavyedge.ProfileData
        Open h5 file of profiles.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Returns
    -------
    (M,) array
        Average profile.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import mean_euclidean
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as f:
    ...     Ys, _, _ = f[:]
    ...     mean = mean_euclidean(f, batch_size=5)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T, "--", color="gray")
    ... plt.plot(mean)
    """
    N, M = f.shape()
    if batch_size is None:
        Ys, _, _ = f[:]
        mean = np.mean(Ys, axis=0, dtype=np.float64)
        logger(f"{N}/{N}")
    else:
        mean = np.zeros((M,), dtype=np.float64)

        for i in range(0, N, batch_size):
            Ys, _, _ = f[i : i + batch_size]
            mean += np.sum(Ys, axis=0)
            logger(f"{i}/{N}")
        mean /= N
    return mean


def mean_wasserstein(f, grid_num, batch_size=None, logger=lambda x: None):
    """Compute mean profile by Fréchet mean with respect to Wasserstein metric.

    Parameters
    ----------
    f : heavyedge.ProfileData
        Open h5 file of profiles.
    grid_num : int
        Number of grids to sample quantile functions.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Returns
    -------
    f_mean : (M,) array
        Average profile.
    L : int
        Length of the support of *f_mean*.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import mean_wasserstein
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as f:
    ...     Ys, _, _ = f[:]
    ...     mean, L = mean_wasserstein(f, 100)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T, "--", color="gray")
    ... plt.plot(mean[:L])
    """
    x = f.x()
    t = np.linspace(0, 1, grid_num)

    N = len(f)
    if batch_size is None:
        Ys, Ls, _ = f[:]
        As = np.trapezoid(Ys, x, axis=-1)
        fs = Ys / As[:, np.newaxis]
        mean, L = wmean(x, fs, Ls, t)
        mean_A = As.mean()
        logger(f"{N}/{N}")
    else:
        g = np.zeros((grid_num,), dtype=np.float64)
        mean_A = 0

        for i in range(0, N, batch_size):
            Ys, Ls, _ = f[i : i + batch_size]
            As = np.trapezoid(Ys, x, axis=-1)
            fs = Ys / As[:, np.newaxis]
            Qs = quantile(x, fs, Ls, t)
            g += np.sum(Qs, axis=0)
            mean_A += np.sum(As)
            logger(f"{i}/{N}")
        g /= N
        mean, L = _wmean(x, t, g)
        mean_A /= N
    return mean * mean_A, L



"""Profile preprocessing."""

import numpy as np

from heavyedge.profile import preprocess

__all__ = [
    "prep",
]


def prep(
    raw_file,
    sigma,
    std_thres,
    fill_value=0.0,
    z_thres=None,
    batch_size=None,
    logger=lambda x: None,
):
    """Preprocess raw profiles in the given file.

    Parameters
    ----------
    raw_file : heavyedge.RawProfileBase
        Opened raw profile file.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.
    std_thres : scalar
        Standard deviation threshold to detect contact point.
    fill_value : scalar, default=0.0
        Value to fill after the contact point.
    z_thres : scalar, optional
        Z-score threshold to detect outliers.
        If not passed, outlier detection is not performed.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Yields
    ------
    Y_processed : (batch_size, M) array
        Preprocessed profiles.
    Ls : (batch_size,) array
        Lengths of the preprocessed profiles.
    names : (batch_size,) array
        Names of the preprocessed profiles.

    Examples
    --------
    >>> from heavyedge import get_sample_path, RawProfileCsvs
    >>> from heavyedge.api import prep
    >>> raw = RawProfileCsvs(get_sample_path("Type3"))
    >>> Ys, Ls, _ = next(prep(raw, 32, 0.01, batch_size=3))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for Y, L in zip(Ys, Ls):
    ...     plt.plot(Y[:L])
    """
    if z_thres is not None:
        gen = _prep_outlier(raw_file, sigma, std_thres, z_thres)
    else:
        gen = _prep(raw_file, sigma, std_thres)

    N = len(raw_file)
    batch_count = 0
    Ys, Ls, names = [], [], []
    for i, (Y, L, name) in enumerate(gen):
        Y[L:] = fill_value
        Ys.append(Y)
        Ls.append(L)
        names.append(name)
        batch_count += 1

        if batch_count == batch_size:
            logger(f"{i}/{N}")
            yield Ys, Ls, names
            Ys, Ls, names = [], [], []
            batch_count = 0

    # yield remaining batch
    logger(f"{N}/{N}")
    yield Ys, Ls, names


def _prep_outlier(raw, sigma, std_thres, z_thres):
    idxs, Ls, sums = [], [], []
    for i in range(len(raw)):
        Y, _ = raw[i]
        if _is_invalid(Y):
            continue
        (Y,), (L,) = preprocess(Y.reshape(1, -1), sigma, std_thres)
        idxs.append(i)
        Ls.append(L)
        sums.append(np.sum(Y[:L]))
    sums = np.array(sums)
    is_outlier = _outlier(sums, z_thres)
    idxs = np.array(idxs)[~is_outlier]

    # yield
    for i, L in zip(idxs, Ls):
        Y, name = raw[i]
        if Y[0] < Y[-1]:
            Y = np.flip(Y)
        yield (Y - Y[L - 1]), L, name


def _outlier(values, thres=3.5):
    # Boris Iglewicz and David C Hoaglin,
    # Volume 16: how to detect and handle outliers. Quality Press, 1993.
    med = np.median(values)
    mad = np.median(np.abs(values - med))
    mod_z = 0.6745 * (values - med) / mad
    return np.abs(mod_z) > thres


def _prep(raw, sigma, std_thres):
    for i in range(len(raw)):
        Y, name = raw[i]
        if _is_invalid(Y):
            continue
        (Y,), (L,) = preprocess(Y.reshape(1, -1), sigma, std_thres)
        yield Y, L, name


def _is_invalid(profile):
    return (len(profile) == 0) or np.any(np.isnan(profile)) or np.any(np.isinf(profile))



"""Raw profile data files."""

import abc
import csv
import warnings
from pathlib import Path

import numpy as np

__all__ = [
    "RawProfileBase",
    "RawProfileCsvs",
]


def _deprecated(version, replace):
    removed_version = str(int(version.split(".")[0]) + 1) + ".0"

    def decorator(func):
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"{func.__name__}() is deprecated since HeavyEdge {version} "
                f"and will be removed in {removed_version}. "
                f"Use {replace} instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


class RawProfileBase(abc.ABC):
    """Base class to read raw profile data.

    All profiles must have the same length.

    Notes
    -----
    ``self[key]`` returns a tuple of profile(s) and profile name(s).
    """

    def __init__(self, path):
        self.path = Path(path).expanduser()

    def __len__(self):
        """Return number of profile data.

        .. note::
            This method will be an abstract method in HeavyEdge 2.0.
            User should implement it with an efficient algorithm.
        """
        return self.count_profiles()

    def __getitem__(self, key):
        """Return profile and name at index.

        .. note::
            This method will be an abstract method in HeavyEdge 2.0.
            User should implement it with an efficient algorithm.
        """
        return (self.all_profiles()[key], np.array(self.profile_names())[key])

    @_deprecated("1.6", "__len__() method")
    @abc.abstractmethod
    def count_profiles(self):
        """Number of raw profiles.

        .. deprecated:: 1.6
            This method will be removed in HeavyEdge 2.0.
            Implement __len__() and use len() instead.

        Returns
        -------
        int
        """

    @_deprecated("1.6", "__getitem__() method")
    def profiles(self):
        """Yield raw profiles.

        .. deprecated:: 1.6
            This method will be removed in HeavyEdge 2.0.
            Implement __getitem__() and index the object instead.

        Yields
        ------
        1-D ndarray
        """

    @_deprecated("1.6", "__getitem__() method")
    @abc.abstractmethod
    def all_profiles(self):
        """Return all profiles as an 2-D array.

        .. deprecated:: 1.6
            This method will be removed in HeavyEdge 2.0.
            Implement __getitem__() and index the object instead.

        Returns
        -------
        2-D ndarray
        """
        return np.array([p for p in self.profiles()])

    @_deprecated("1.6", "__getitem__() method")
    @abc.abstractmethod
    def profile_names(self):
        """Yield profile names.

        .. deprecated:: 1.6
            This method will be removed in HeavyEdge 2.0.
            Implement __getitem__() and index the object instead.

        Yields
        ------
        str
        """


class RawProfileCsvs(RawProfileBase):
    """Read raw profile data from a directory containing CSV files.

    Directory structure:

    .. code-block::

        rawdata/
        ├── profile1.csv
        ├── profile2.csv
        └── ...

    Parameters
    ----------
    path : pathlike
        Path to the directory containing the raw CSV files.

    Notes
    -----
    - Each CSV file must contain a single column of numeric values (no header).
    - The order of profiles is determined by the sorted filenames.
    - The profile name is derived from the filename stem.

    Examples
    --------
    >>> from heavyedge import get_sample_path, RawProfileCsvs
    >>> profiles = RawProfileCsvs(get_sample_path("Type3"))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for i in range(len(profiles)):
    ...     profile, _ = profiles[i]
    ...     plt.plot(profile)
    """

    def __init__(self, path):
        super().__init__(path)
        self._files = sorted(self.path.glob("*.csv"))

    def __len__(self):
        return len(self._files)

    @staticmethod
    def _read_profile(path):
        with open(path, newline="") as csvfile:
            reader = csv.reader(csvfile)
            profile = np.array([float(row[0]) for row in reader])
        return profile

    def __getitem__(self, key):
        file = self._files[key]
        return (self._read_profile(file), str(file.stem))

    def count_profiles(self):
        # TODO: remove in HeavyEdge 2.0
        return len(self)

    def profiles(self):
        # TODO: remove in HeavyEdge 2.0
        for file in self._files:
            yield self._read_profile(file)

    def profile_names(self):
        # TODO: remove in HeavyEdge 2.0
        for f in self._files:
            yield str(f.stem)



"""Commands to average profiles."""

import pathlib

from heavyedge.cli.command import Command, register_command

PLUGIN_ORDER = 0.1


@register_command("mean", "Compute mean profile")
class MeanCommand(Command):
    def add_parser(self, main_parser):
        mean = main_parser.add_parser(
            self.name,
            description="Compute Wasserstein mean profile and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        mean.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        mean.add_config_argument(
            "--wnum",
            type=int,
            help="Number of sample points to compute Wasserstein mean.",
        )
        mean.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, loads entire profiles.",
        )
        mean.add_argument(
            "-o", "--output", type=pathlib.Path, help="Output npy file path"
        )

    def run(self, args):
        import numpy as np

        from heavyedge import ProfileData
        from heavyedge.api import mean_wasserstein

        self.logger.info(f"Writing {args.output}")

        with ProfileData(args.profiles) as file:
            _, M = file.shape()
            res = file.resolution()
            name = file.name()

            with ProfileData(args.output, "w").create(M, res, name) as out:
                mean, L = mean_wasserstein(
                    file,
                    args.wnum,
                    args.batch_size,
                    lambda msg: self.logger.info(f"{out.path} : {msg}"),
                )
                mean[L:] = np.nan

                out.write_profiles(mean.reshape(1, -1), [L], [name])

        self.logger.info(f"Saved {out.path}.")



"""High-level Python runtime interface."""

__all__ = [
    "prep",
    "preprocess",
    "fill_after",
    "outlier",
    "mean",
    "mean_euclidean",
    "mean_wasserstein",
    "landmarks_type2",
    "landmarks_type3",
    "plateau_type2",
    "plateau_type3",
    "scale_area",
    "scale_plateau",
    "trim",
    "pad",
]

from .edge import pad, scale_area, scale_plateau, trim
from .landmarks import landmarks_type2, landmarks_type3, plateau_type2, plateau_type3
from .mean import mean_euclidean, mean_wasserstein
from .preprocess import prep
from .profile import fill_after, mean, outlier, preprocess



"""Edge manipulation."""

import numpy as np

from heavyedge.profile import fill_after

__all__ = [
    "scale_area",
    "scale_plateau",
    "trim",
    "pad",
]


def scale_area(f, batch_size=None, logger=lambda x: None):
    """Scale edge profile by area.

    Parameters
    ----------
    f : heavyedge.ProfileData
        Open h5 file of profiles.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Yields
    ------
    scaled : (batch_size, M) array
        Scaled edge profile.
    Ls : (batch_size,) array
        Lengths of the scaled profiles.
    names : (batch_size,) array
        Names of the scaled profiles.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import scale_area
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as f:
    ...     gen = scale_area(f, batch_size=5)
    ...     Ys = np.concatenate([ys for ys, _, _ in gen], axis=0)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T)
    """
    x = f.x()

    N = len(f)
    if batch_size is None:
        Ys, Ls, names = f[:]
        Ys /= _area(x, Ys, Ls)[:, np.newaxis]
        logger(f"{N}/{N}")
        yield Ys, Ls, names
    else:
        for i in range(0, N, batch_size):
            Ys, Ls, names = f[i : i + batch_size]
            Ys /= _area(x, Ys, Ls)[:, np.newaxis]
            logger(f"{i}/{N}")
            yield Ys, Ls, names


def _area(x, Ys, Ls):
    Ys = Ys.copy()
    fill_after(Ys, Ls, 0)
    return np.trapezoid(Ys, x, axis=1)


def scale_plateau(f, batch_size=None, logger=lambda x: None):
    """Scale edge profile by plateau height.

    Parameters
    ----------
    f : heavyedge.ProfileData
        Open h5 file of profiles.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Yields
    ------
    scaled : (batch_size, M) array
        Scaled edge profile.
    Ls : (batch_size,) array
        Lengths of the scaled profiles.
    names : (batch_size,) array
        Names of the scaled profiles.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import scale_plateau
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as f:
    ...     gen = scale_plateau(f, batch_size=5)
    ...     Ys = np.concatenate([ys for ys, _, _ in gen], axis=0)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T)
    """
    N = len(f)
    if batch_size is None:
        Ys, Ls, names = f[:]
        Ys /= Ys[:, [0]]
        logger(f"{N}/{N}")
        yield Ys, Ls, names
    else:
        for i in range(0, N, batch_size):
            Ys, Ls, names = f[i : i + batch_size]
            Ys /= Ys[:, [0]]
            logger(f"{i}/{N}")
            yield Ys, Ls, names


def trim(f, width1, width2, batch_size=None, logger=lambda x: None):
    """Trim edge profile to a specific width.

    This function matches the contact points of all profiles to a same location.

    Parameters
    ----------
    f : heavyedge.ProfileData
    width1 : int
        Number of points on the left side of the profile.
    width2 : int
        Number of points on the right side of the profile.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Yields
    ------
    trimmed : (batch_size, width) array
        Trimmed edge profile.
    Ls : (batch_size,) array
        Lengths of the scaled profiles.
    names : (batch_size,) array
        Names of the scaled profiles.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import trim
    >>> with ProfileData(get_sample_path("MeanProfiles.h5")) as f:
    ...     Ys, _, _ = f[:]
    ...     gen = trim(f, 1500, 0, batch_size=10)
    ...     Ys_trim = np.concatenate([ys for ys, _, _ in gen], axis=0)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T)
    ... plt.plot(Ys_trim.T)
    """
    N, _ = f.shape()
    Ls = f._file["len"][:]

    if batch_size is None:
        Ys, Ls, names = f[:]
        logger(f"{N}/{N}")
        yield _trim(Ys, Ls, width1, width2), Ls, names
    else:
        for i in range(0, N, batch_size):
            Ys, Ls, names = f[i : i + batch_size]
            logger(f"{i}/{N}")
            yield _trim(Ys, Ls, width1, width2), Ls, names


def _trim(Ys, Ls, w1, w2):
    N, M = Ys.shape
    idx_array = np.arange(M)[None, :]
    mask1 = idx_array >= (Ls - w1)[:, None]
    mask2 = idx_array < (Ls + w2)[:, None]
    ret = Ys[mask1 & mask2].reshape(N, w1 + w2)
    return ret


def pad(f, width1, width2, batch_size=None, logger=lambda x: None):
    """Pad edge profile to a specific width.

    Parameters
    ----------
    f : heavyedge.ProfileData
    width1 : int
        Number of points on the left side of the profile.
    width2 : int
        Number of points on the right side of the profile.
    batch_size : int, optional
        Batch size to load data.
        If not passed, all data are loaded at once.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Yields
    ------
    padded : (batch_size, width) array
        Padded edge profile.
    Ls : (batch_size,) array
        Lengths of the padded profiles.
    names : (batch_size,) array
        Names of the padded profiles.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import pad
    >>> with ProfileData(get_sample_path("MeanProfiles.h5")) as f:
    ...     Ys, _, _ = f[:]
    ...     gen = pad(f, 5000, 100, batch_size=10)
    ...     Ys_pad = np.concatenate([ys for ys, _, _ in gen], axis=0)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Ys.T)
    ... plt.plot(Ys_pad.T)
    """
    N, _ = f.shape()
    Ls = f._file["len"][:]

    if batch_size is None:
        Ys, Ls, names = f[:]
        logger(f"{N}/{N}")
        yield _pad(Ys, Ls, width1, width2), Ls, names
    else:
        for i in range(0, N, batch_size):
            Ys, Ls, names = f[i : i + batch_size]
            logger(f"{i}/{N}")
            yield _pad(Ys, Ls, width1, width2), Ls, names


def _pad(Ys, Ls, w1, w2):
    N, M = Ys.shape
    ret = np.empty((N, w1 + w2), dtype=Ys.dtype)

    ret[:w1] = Ys[:, :1]
    ret_mask = np.arange(w1 + w2)[None, :] >= (w1 - Ls)[:, None]
    Ys_mask = np.arange(M)[None, :] < (w2 + Ls)[:, None]
    ret[ret_mask] = Ys[Ys_mask]
    return ret



"""Commands for edge data manipulation."""

import pathlib

from heavyedge.cli.command import Command, register_command

PLUGIN_ORDER = 0.2


@register_command("scale", "Scale edge profiles")
class ScaleCommand(Command):
    def add_parser(self, main_parser):
        scale = main_parser.add_parser(
            self.name,
            description="Scale edge profiles.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        scale.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        scale.add_argument(
            "--type",
            choices=["area", "plateau"],
            default="area",
            help="Scaling type (default=area).",
        )
        scale.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, loads entire profiles.",
        )
        scale.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.api import scale_area, scale_plateau
        from heavyedge.io import ProfileData

        if args.type == "area":
            scale = scale_area
        elif args.type == "plateau":
            scale = scale_plateau
        else:
            raise NotImplementedError

        self.logger.info(f"Writing {args.output}")

        with ProfileData(args.profiles) as file:
            _, M = file.shape()
            res = file.resolution()
            name = file.name()

            with ProfileData(args.output, "w").create(M, res, name) as out:

                for scaled, Ls, names in scale(
                    file,
                    args.batch_size,
                    lambda msg: self.logger.info(f"{out.path} : {msg}"),
                ):
                    out.write_profiles(scaled, Ls, names)

        self.logger.info(f"Saved {out.path}.")


@register_command("trim", "Trim edge profiles")
class TrimCommand(Command):
    def add_parser(self, main_parser):
        trim = main_parser.add_parser(
            self.name,
            description="Trim edge profiles by a fixed width.",
            epilog=(
                "Width unit is determined by the resolution of 'profile'. "
                "The resulting hdf5 file is in 'ProfileData' structure."
            ),
        )
        trim.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        trim.add_config_argument(
            "--width",
            type=float,
            help="Edge width. If not passed, length of the shortest profile.",
        )
        trim.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, loads entire profiles.",
        )
        trim.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.api import trim
        from heavyedge.io import ProfileData

        self.logger.info(f"Writing {args.output}")

        with ProfileData(args.profiles) as file:
            _, M = file.shape()
            res = file.resolution()
            name = file.name()

            Ls = file._file["len"][:]
            if args.width is None:
                args.width = Ls.min() / res

            w1 = int(args.width * res)
            w2 = (M - Ls).min()
            with ProfileData(args.output, "w").create(w1 + w2, res, name) as out:

                for trimmed, Ls, names in trim(
                    file,
                    w1,
                    w2,
                    args.batch_size,
                    lambda msg: self.logger.info(f"{out.path} : {msg}"),
                ):
                    out.write_profiles(trimmed, Ls, names)

        self.logger.info(f"Saved {out.path}.")


@register_command("pad", "Pad edge profiles")
class PadCommand(Command):
    def add_parser(self, main_parser):
        pad = main_parser.add_parser(
            self.name,
            description="Pad edge profiles to a fixed width.",
            epilog=(
                "Width unit is determined by the resolution of 'profile'. "
                "The resulting hdf5 file is in 'ProfileData' structure."
            ),
        )
        pad.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        pad.add_config_argument(
            "--width",
            type=float,
            help="Edge width. If not passed, length of the shortest profile.",
        )
        pad.add_argument(
            "--batch-size",
            type=int,
            help="Batch size to load data. If not provided, loads entire profiles.",
        )
        pad.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.api import pad
        from heavyedge.io import ProfileData

        self.logger.info(f"Writing {args.output}")

        with ProfileData(args.profiles) as file:
            _, M = file.shape()
            res = file.resolution()
            name = file.name()

            Ls = file._file["len"][:]
            if args.width is None:
                args.width = Ls.max() / res

            w1 = int(args.width * res)
            w2 = (M - Ls).min()
            with ProfileData(args.output, "w").create(w1 + w2, res, name) as out:

                for padded, Ls, names in pad(
                    file,
                    w1,
                    w2,
                    args.batch_size,
                    lambda msg: self.logger.info(f"{out.path} : {msg}"),
                ):
                    out.write_profiles(padded, Ls, names)

        self.logger.info(f"Saved {out.path}.")
